/**
 * IBM Confidential OCO Source Material
 * 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2011
 * The source code for this program is not published or otherwise divested
 * of its trade secrets, irrespective of what has been deposited with the
 * U.S. Copyright Office.
 *
 * Change History:
 * ----------------------------------------------------------------------------
 * $F36603                2011.03.17  ralder    birth
 * ----------------------------------------------------------------------------
 */

package com.ibm.websphere.batch.devframework.datastreams.bdsadapter;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.Serializable;
import java.util.HashMap;
import java.util.concurrent.TimeUnit;

import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.SkipListener;
import com.ibm.websphere.batch.RecordMetrics;
import com.ibm.websphere.batch.RecordMetrics.MetricName;
import com.ibm.websphere.batch.context.JobStepContextMgr;
import com.ibm.ws.gridcontainer.batch.impl.SkipHandler;

/**
 * Base class provides skipped-record handling. 
 * @ibm-api
 */
public abstract class AbstractBatchDataInputStreamRecordMetrics
  extends AbstractBatchDataStream
  implements AbstractBatchDataInputStream, AbstractBatchDataStreamRecordMetrics, RecordMetrics, Externalizable
{

  /** Manages the processing of skipped records */
  private SkipHandler _skipHandler = null;
  
  private HashMap<MetricName,Long> metrics = new HashMap<MetricName,Long>(2);
  private long startTime; 
  private long _recordCount = 0;
  private long _cumulativeTime = 0;
  private int _checkpointSkipCount = 0;


  public void initialize(String name, String jobstepId) throws BatchContainerDataStreamException 
  {
    super.initialize(name, jobstepId);

    String[] jobstep = jobstepId.split("/",2);
    _skipHandler = new SkipHandler(this, jobstep[0], jobstep[1]);
    if (JobStepContextMgr.getContext().getRecordMetrics(this.getName()) == null)
    		JobStepContextMgr.getContext().setRecordMetrics(this.getName(), this);
    else {
    	_recordCount = ((AbstractBatchDataInputStreamRecordMetrics)JobStepContextMgr.getContext().getRecordMetrics(this.getName())).getRecordCount();
    	_cumulativeTime = ((AbstractBatchDataInputStreamRecordMetrics)JobStepContextMgr.getContext().getRecordMetrics(this.getName())).getCumulativeTime();
    }    	
  }


  /**
   * Abstract method implemented by sub class.  This method is called by hasNext.
   */
  protected abstract boolean _hasNext() throws Exception;

	/**
	 * Return a boolean indicating if there is unread data in the Input Stream.
   * Delegates to _hasNext, implemented by sub class.
   *
   * Handle skip records, if applicable.
   *
	 * @return boolean indicating if there is more data to be read
	 * @throws Exception
	 */
	public final boolean hasNext() throws Exception
  {
    while (true)  
    {
      // Loop for skip records.  Either the code will return successfully or 
      // throw an exception or skip and loop around (hence, the infinite while loop).
      try
      {
        return _hasNext();
      }
      catch (Exception e)
      {
        // SkipHandler.handleException will re-throw the exception if it cannot be skipped.
        _skipHandler.handleException(e);
        setSkipMetric();
      }
    }
  }
	

  /**
   * Abstract method implemented by sub class.  This method is called by read().
   */
  protected abstract Object _read() throws Exception;


	/**
	 * Return the next record.
   * Delegates to _read, implemented by the sub class.
   *
   * Handle skip records, if applicable.
   *
	 * @return an object representing the next record in the inputstream
	 * @throws Exception
	 */
	public final Object read() throws Exception
  {
    while (true)  
    {
      // Loop for skip records.  Either the code will return successfully or 
      // throw an exception or skip and loop around (hence, the infinite while loop).
      try
      {
    	startRPSMetric();  
        Object record = _read();
        endRPSMetric();
        return record;
      }
      catch (Exception e)
      {
        // SkipHandler.handleException will re-throw the exception if it cannot be skipped.
        _skipHandler.handleException(e);
        setSkipMetric();
      }
    }
    
  }

	
  /**
   * Abstract method implemented by sub class.  This method is called by
   * read(Object).
   */
  protected abstract Object _read(Object parameters) throws Exception;


	/**
	 * Return the next record based on the input parameters.
   * Delegates to _read(Object), implemented by the sub class.
   *
   * Handle skip records, if applicable.
   *
	 * @param parameters
	 * @return
	 * @throws Exception
	 */
	public final Object read(Object parameters) throws Exception
  {
    while (true)  
    {
      // Loop for skip records.  Either the code will return successfully or 
      // throw an exception or skip and loop around (hence, the infinite while loop).
      try
      {
    	startRPSMetric();  
        Object record = _read(parameters);
        endRPSMetric();
        return record;
      }
      catch (Exception e)
      {
        // SkipHandler.handleException will re-throw the exception if it cannot be skipped.
        _skipHandler.handleException(e);
        setSkipMetric();
      }
    }
  }
	
	public abstract Object fetchHeader();


  /**
   * Abstract method implemented by sub class.  This method is called 
   * by internalizeCheckpointInformation.
   *
   * @param token the checkpoint token
   */
	public abstract void _internalizeCheckpointInformation(String token) ;

	/**
	 * Parse record metric data from the beginning of the token, then call the sub 
   * class (via _internalizeCheckpointInformation), passing the rest of the token.
	 * 
   * @param token the checkpoint token
	 */
	public final void internalizeCheckpointInformation(String token) 
  {
    String[] tokens = token.split(";",2);

    _checkpointSkipCount = Integer.parseInt(tokens[0]);
    _internalizeCheckpointInformation(tokens[1]);
  }


  /**
   * Abstract method implemented by sub class.  This method is called by
   * externalizeCheckpointInformation.
   */
	public abstract String _externalizeCheckpointInformation();

	/**
   * Encode record metric data at the beginning of the token, then call the sub class 
   * (via _externalizeCheckpointInformation) and append whatever it returns to the 
   * end of the token.
   *
	 */
	public final String externalizeCheckpointInformation()
  {
    String token = Long.toString(_skipHandler.getSkipCount()) + ";";

    String subClassToken = _externalizeCheckpointInformation();
    if (subClassToken != null)
      token += subClassToken;

    return token;
  }


  /**
   * Abstract method implemented by sub class.  This method is called by
   * positionAtCurrentCheckpoint.
   */
  public abstract void _positionAtCurrentCheckpoint() throws BatchContainerDataStreamException;

  /**
   * This method calls the sub class (via _positionAtCurrentCheckpoint) to allow the sub class 
   * to position the BDS properly, then it applies checkpoint-related record metric settings
   * to the BDS.
   */
  public void positionAtCurrentCheckpoint() throws BatchContainerDataStreamException
  {
    _positionAtCurrentCheckpoint(); // call sub class to do it's thing.

    // d707776 - Set the skip count after the sub class processes the checkpoint.
    // Why?  Because as the sub class repositions itself to the checkpoint,
    // it may inadvertantly recreate all the skip errors that occurred 
    // leading up to the checkpoint.   We don't want the recreated skip errors
    // to be counted ON TOP OF the persisted skip count (which alredy includes
    // those errors).  So essentialy we're trying to avoid double-bookkeeping 
    // of the skips.
    _skipHandler.setSkipCount(_checkpointSkipCount);
  }


  /**
   * <p>
   * Add a SkipListener callback to the BDS.  The SkipListener is called by the BDS whenever the
   * BDS fails to read or write a record and is configured to skip the failed record.
   *
   * @param skipListener The user-implemented SkipListener callback.  
   */
  public void addSkipListener(SkipListener skipListener)
  {
    _skipHandler.addSkipListener(skipListener);
  }
  
  public long getMetric(MetricName metricName) {
	  if (metrics.get(metricName) != null)
		  return metrics.get(metricName).longValue();
	  else
		  return 0;
  }
  
  public long getRecordCount() {
	return _recordCount;
  }
  
  public long getCumulativeTime() {
	return _cumulativeTime;
  }
		
  private void setMetric(MetricName metricName, long value) {
	metrics.put(metricName,Long.valueOf(value));
  }
  
  protected void startRPSMetric() {
	startTime = System.nanoTime();
  }
  
  protected void endRPSMetric() {
	long endTime = System.nanoTime();
	_cumulativeTime += (endTime - startTime);
	++_recordCount;	
	if (_cumulativeTime > 0) {
		long rps = 1000000000/(_cumulativeTime/_recordCount);
		setMetric(MetricName.rps, rps);
		JobStepContextMgr.getContext().setRecordMetrics(getName(), this);
	}
  }
  
  protected void setSkipMetric() {
	setMetric(MetricName.skip, _skipHandler.getSkipCount());
	JobStepContextMgr.getContext().setRecordMetrics(getName(), this);
  }
  
  public void readExternal(ObjectInput in) throws IOException,
	ClassNotFoundException {
	  metrics = (HashMap)in.readObject();
	  _recordCount = in.readLong();
	  _cumulativeTime = in.readLong();
  }

  public void writeExternal(ObjectOutput out) throws IOException {
	  out.writeObject(metrics);
	  out.writeLong(_recordCount);
	  out.writeLong(_cumulativeTime);
  }

}
