// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008
/**
 * @author Antani / Neeraj
 * This class is responsible for reading from a Byte stream
 */

package com.ibm.websphere.batch.devframework.datastreams.patternadapter;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.util.Properties;

import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;
import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.PerformanceAnalyzer;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataInputStreamRecordMetrics;

public abstract class BDSFileByteReader extends AbstractBatchDataInputStreamRecordMetrics
{

	/**
	 * This class handles the basic book-keeping related to 
	 * reading a Byte stream and delegates the "business logic"
	 * to the implementer of ByteReaderPattern
	 */
	protected ByteReaderPattern byteReaderImpl; 
	
	/**
	 * Holds the byte input stream 
	 */
	protected BufferedInputStream bufferedByteReader; 
	
  /**
   * Indicates that a record has been prefetched and is waiting to be read.
   */
	protected boolean prefetchedRecordAvailable = false;
	
	/**
	 * basic unit of processing
	 */
	protected Object record;
	
	/**
	 * Name of the file to be read
	 */
	protected String fileName = "";
	
	
	protected int recordCount;
	
	/**
	 * Holds the properties that the user provides in the xJCL
	 */
	protected Properties props;	
	
	protected BDSFWLogger logger;
	
	/**
	 * Holds the performance metrics for this class
	 */
	protected PerformanceAnalyzer perfAnalyzer;
	
	protected boolean isPerformanceMeasurementEnabled = false;
	
	private static final String  CLASSNAME = BDSFileByteReader.class.getName();
	
	public abstract void open() throws BatchContainerDataStreamException;
	
	/**
	 * Store the properties passed through the xJCL
	 * Instantiate the logger
	 * Instantiate the bytereader class that the user implements
	 * 
	 */
	public void initialize(Properties props)
	{
		logger = new BDSFWLogger(props);
		perfAnalyzer = new PerformanceAnalyzer(CLASSNAME, props);
		
		isPerformanceMeasurementEnabled = perfAnalyzer.isEnabled();
		
		try 
		{	
			
			this.props = props;
			
			String implClass = props.getProperty(BDSFrameworkConstants.IMPLCLASS_KEY);
			// For backward compatibility
			if(implClass == null)
				implClass = getRequiredProperty(props,"IMPLCLASS");
			byteReaderImpl = (ByteReaderPattern) loadClass(implClass);
			if(logger.isDebugEnabled())
				logger.info("Succesfully initialized" + implClass);
			byteReaderImpl.initialize(props);				
		}
		catch (Throwable t)
		{
			throw new RuntimeException(t);
		}
		
	}
	
	/**
	 * This method delegates retrieval of the next record to the user provided
	 * pattern implementation class
	 */
	public Object _read() throws IOException {
		String methodName = "read";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		Object returnValue = null;

    // d707776.1: If a prefetched record is available, return it (and reset
    // the prefetchedRecordAvailable flag).  Otherwise, try to fetch a record.
    if (prefetchedRecordAvailable)
    {
      returnValue = record;
      prefetchedRecordAvailable = false;
    }
    else
    {
      // d707776.1: Increment recordCount prior to fetch.  We don't want recordCount
      // influenced by skippable exceptions that might get thrown by fetchRecord.
      // This is because recordCount is used to reposition the stream after a checkpoint, 
      // so we want to know exactly how many records we *attempted* to read, regardless 
      // of how many were skipped, so we can reposition the stream accurately.
		  recordCount++;		
			returnValue = byteReaderImpl.fetchRecord(bufferedByteReader);
    }

		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		
		return returnValue;
	}
	
	/**
	 * This method is not supported
	 */
	
	
	public Object _read(Object parameters) throws Exception {
		throw new Exception("Unsupported");
		
	}
	
	/**
	 * Returns true if there is more data to be read
	 */

	public boolean _hasNext() throws Exception {
		String methodName = "hasNext";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME + ":" + methodName);

    // d707776.1: If a prefetched record is available -- meaning it hasn't
    // yet been read via the read() method -- then don't do anything here.  
    // Just return true, informing the caller that the stream has more records.  
    // Otherwise, prefetch the next record.
    if (!prefetchedRecordAvailable && bufferedByteReader.available() != 0)
    {
      // No prefetch record available, but the stream still has data.  
      // Prefetch the next record.
      record = read();
		  if (record != null)
        prefetchedRecordAvailable = true;
    }
    
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		return prefetchedRecordAvailable;
	}
	
	/**
	 * Close the stream
	 */

	public void close() throws BatchContainerDataStreamException {

		try {				
			bufferedByteReader.close();
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.displayPerformanceStatisticsForEntity();
		} catch (Throwable t) {
			throw new BatchContainerDataStreamException(t);
		}
	}
	
	/**
	 * Return the current recordcount
	 */
	
	public String _externalizeCheckpointInformation() {
		
		if (logger.isDebugEnabled())
			logger.debug(CLASSNAME  + ".externalizeCheckpointInfo- token= |" + recordCount + "|");
		
		return String.valueOf(recordCount);		
	}
	
	/**
	 * Position at the given record number
	 */
	
	public void _internalizeCheckpointInformation(String token)
			throws RuntimeException {
		String methodName = "internalizeCheckpointInformation";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		if (logger.isDebugEnabled())
			logger.debug("Entering " + CLASSNAME + "." + methodName +"- token= |" + token + "|");
		
		try {
			if (token == null)
				positionAtInitialCheckpoint();

			else
				recordCount = Integer.valueOf(token).intValue();
		} catch (Throwable t) {
			throw new RuntimeException(t);
		}
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		if (logger.isDebugEnabled())
			logger.debug("Exiting " + CLASSNAME + "." + methodName + "- token= |" + token + "|");
		
	}
	
	/**
	 * Invoked when the stream is first opened. 
	 */
	
	public void positionAtInitialCheckpoint()
			throws BatchContainerDataStreamException {
		recordCount = 0;
	}
	
	/**
	 * Start from the beginning and read until the current record = recordcount
	 */
	
	public void _positionAtCurrentCheckpoint()
			throws BatchContainerDataStreamException {			
		String methodName = "positionAtCurrentCheckpoint";
		if(logger.isDebugEnabled())
			logger.debug("Entering " + CLASSNAME + ":" + methodName);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		setRestart(true);
		int repositionCount = recordCount;
		recordCount = 0;
		try {			
      // d707776.1: keep calling read() until recordCount hits the repositionCount.
      // We don't just want to call read() *repositionCount* times, because some of
      // the reads might incur skippable exceptions, which will cause the underlying
      // framework to automatically read() again, under the covers.  That additional
      // read wouldn't get counted here, so we'd end up calling read() too many times.
      // Instead, just keep calling read(), which will update recordCount on its own,
      // and wait until recordCount matches the checkpoint value (repositionCount).
			while (recordCount < repositionCount)
			{
				read();
			}
			
		} catch (Exception e) {
			throw new BatchContainerDataStreamException(e);
		}
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		if(logger.isDebugEnabled())
			logger.debug("Exiting " + CLASSNAME + ":"  + methodName);
		
	}
	
	/**
	 * Invokes the user's implementation of the ByteReader pattern to
	 * retrieve the file header.
	 */
	public Object fetchHeader() {
		Object header = null;
		String methodName = "fetchHeader";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		
		header = byteReaderImpl.fetchHeader();
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		return header;
	}
}
