// Licensed Materials - Property of IBM 
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008

/**
 * @author Antani/Neeraj
 * This class is responsible for reading from a text stream
 */
package com.ibm.websphere.batch.devframework.datastreams.patternadapter;

import java.io.BufferedReader;
import java.io.IOException;
import java.util.Properties;

import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.devframework.configuration.BDSFrameworkConstants;
import com.ibm.websphere.batch.devframework.configuration.BDSFWLogger;
import com.ibm.websphere.batch.devframework.configuration.PerformanceAnalyzer;
import com.ibm.websphere.batch.devframework.datastreams.bdsadapter.AbstractBatchDataInputStreamRecordMetrics;
import com.ibm.websphere.batch.devframework.datastreams.patternadapter.FileReaderPattern;

public abstract class BDSFileReader extends AbstractBatchDataInputStreamRecordMetrics {

	public static final String CLASSNAME = "BDSFileReader";
	protected int recordCount;

	protected FileReaderPattern fileReaderImpl;

  /**
   * Indicates that a record has been prefetched and is waiting to be read.
   */
  protected boolean prefetchedRecordAvailable = false;

	protected Object record;

	protected BufferedReader reader;

	protected String fileName = "";
	
	protected Properties props;	
	
	protected PerformanceAnalyzer perfAnalyzer = null;
	protected boolean isPerformanceMeasurementEnabled = false;
	protected BDSFWLogger logger;
	
	
	protected String fileEncoding = null;

	/**
	 * Invoked during stream initialization. This method loads the user provided
	 * pattern implementation class
	 */
	public void initialize(Properties props)
	{
		try 
		{
			logger = new BDSFWLogger(props);
			perfAnalyzer = new PerformanceAnalyzer(CLASSNAME, props);
			isPerformanceMeasurementEnabled = perfAnalyzer.isEnabled();
			this.props = props;
			fileEncoding = props.getProperty(BDSFrameworkConstants.FILE_ENCODING_KEY);
			String fileReaderImplClass = props.getProperty(BDSFrameworkConstants.IMPLCLASS_KEY);
			// For backward compatibility
			if(fileReaderImplClass == null)
				fileReaderImplClass = getRequiredProperty(props,"IMPLCLASS");			
			fileReaderImpl = (FileReaderPattern) loadClass(fileReaderImplClass);
			if (logger.isDebugEnabled())
				logger.debug("Loaded filereaderpattern" + fileReaderImplClass);
			fileReaderImpl.initialize(props);
			
		}
		catch (Throwable t)
		{
			logger.error("Exception occured loading filereaderpattern" + t.getMessage());
			throw new RuntimeException(t);
		}
	}
	
	public abstract void open() throws BatchContainerDataStreamException;
	
	public void _positionAtCurrentCheckpoint()
			throws BatchContainerDataStreamException {
		String methodName = "positionAtCurrentCheckpoint";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		setRestart(true);
		int repositionCount = recordCount;
		recordCount = 0;
		try {
      // d707776.1: keep calling read() until recordCount hits the repositionCount.
      // We don't just want to call read() *repositionCount* times, because some of
      // the reads might incur skippable exceptions, which will cause the underlying
      // framework to automatically read() again, under the covers.  That additional
      // read wouldn't get counted here, so we'd end up calling read() too many times.
      // Instead, just keep calling read(), which will update recordCount on its own,
      // and wait until recordCount matches the checkpoint value (repositionCount).
			while (recordCount < repositionCount) {
				read();
			}
		} catch (Exception e) {
			throw new BatchContainerDataStreamException(e);
		}
		if(isPerformanceMeasurementEnabled) 
			perfAnalyzer.endMeasurementAndSave(methodName);
	}

	public Object _read() throws IOException {
		String methodName = "read";
		if (logger.isDebugEnabled())
			logger.debug("Entering " + CLASSNAME + ":" + methodName);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		Object returnValue = null;

    // d707776.1: If a prefetched record is available, return it (and reset
    // the prefetchedRecordAvailable flag).  Otherwise, try to fetch a record.
    if (prefetchedRecordAvailable)
    {
      returnValue = record;
      prefetchedRecordAvailable = false;
    }
    else
    {
      // d707776.1: Increment recordCount prior to fetch.  We don't want recordCount
      // influenced by skippable exceptions that might get thrown by fetchRecord.
      // This is because recordCount is used to reposition the stream after a checkpoint, 
      // so we want to know exactly how many records we *attempted* to read, regardless 
      // of how many were skipped, so we can reposition the stream accurately.
		  recordCount++;		
			returnValue = fileReaderImpl.fetchRecord(reader);
    }

		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		return returnValue;
	}
	
	
	public Object _read(Object parameters) throws Exception {
		throw new Exception("Unsupported");
		
	}

	public boolean _hasNext() throws Exception {
		String methodName = "hasNext";
		if(logger.isDebugEnabled())
			logger.debug("Entering " + CLASSNAME + "." + methodName);
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);

    // d707776.1: If a prefetched record is available -- meaning it hasn't
    // yet been read via the read() method -- then don't do anything here.  
    // Just return true, informing the caller that the stream has more records.  
    // Otherwise, prefetch the next record.
    if (!prefetchedRecordAvailable)
    {
      // No prefetch record available.  Go fetch one to see if the stream is still live.
      record = read();
		  if (record != null)
        prefetchedRecordAvailable = true;
    }

		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		if(logger.isDebugEnabled())
			logger.debug("Exiting " + CLASSNAME + "." + methodName + " with " + prefetchedRecordAvailable);
		return prefetchedRecordAvailable;
	}

	public void close() throws BatchContainerDataStreamException {

		try {
			reader.close();
			if(isPerformanceMeasurementEnabled)
				perfAnalyzer.displayPerformanceStatisticsForEntity();
		} catch (Throwable t) {
			throw new BatchContainerDataStreamException(t);
		}
	}

	public String _externalizeCheckpointInformation() {
		
		if (logger.isDebugEnabled())
			logger.debug("BDSFileReaderr.externalizeCheckpointInfo- token= |" + recordCount + "|");
		
		return String.valueOf(recordCount);
	}
	
	public void _internalizeCheckpointInformation(String token)
			throws RuntimeException {
		String methodName = "internalizeCheckpointInformation";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		try {
			if (logger.isDebugEnabled())
				logger.debug(CLASSNAME +"." + methodName + "-  token= |" + token + "|");
			
			if (token == null)
				positionAtInitialCheckpoint();
			else			
				recordCount = Integer.parseInt(token);
		} catch (Throwable t) {
			throw new RuntimeException(t);
		}
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
	}
	
	public void positionAtInitialCheckpoint()
			throws BatchContainerDataStreamException {
		recordCount = 0;
	}
	
	public Object fetchHeader() {
		String methodName = "fetchHeader";
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.startMeasurement(methodName);
		Object header = null;
		header = fileReaderImpl.fetchHeader();
		if(isPerformanceMeasurementEnabled)
			perfAnalyzer.endMeasurementAndSave(methodName);
		return header;
	}
}
